// $Id: J4VDetectorComponent.cc,v 1.1.1.1 2001/08/31 07:38:35 hoshina Exp $
//*************************************************************************
//* --------------------
//* J4VDetectorComponent
//* --------------------
//* (Description)
//* 	Class for describing his/her detector compornents.
//*     
//* (Update Record)
//*	2000/12/08  K.Hoshina	Original version.
//*************************************************************************

#include "geomdefs.hh"
#include "G4RotationMatrix.hh"
#include "J4VDetectorComponent.hh"
#include "J4UnionSolid.hh"

//************************************************************************
//---------------------
// Class Description
//---------------------

//=====================================================================
//* default constructor -----------------------------------------------


J4VDetectorComponent::J4VDetectorComponent(G4String& groupname,
					   G4String& name,
					   J4VDetectorComponent* parent,
					   G4int     nclones,
					   G4int     nbrothers,
					   G4int     me) :
		      fSubGroup(groupname), fName(name),
		      fMother(parent), fIsOn(FALSE), 
		      fInstalled(FALSE), fSolid(0), fLV(0), fPV(0), 
		      fSD(0), fNClones(nclones),
		      fNBrothers(nbrothers), fMyID(me), 
		      fCanDeleteUserLimits(FALSE)
{
   if (nclones < 1) {
     G4cerr<<"J4VDetectorComponent::constructor: No of clones must bigger than 0."<<G4endl;
   }
   if (nbrothers < 1) {	
     G4cerr<<"J4VDetectorComponent::constructor: No of brothers must bigger than 0."<<G4endl;
   }
	
   OpenMaterialStore();	
   fSDManager= G4SDManager::GetSDMpointer();
      
   SetName(name, nbrothers, me, parent);
   
}

//=====================================================================
//* destructor --------------------------------------------------------

J4VDetectorComponent::~J4VDetectorComponent()
{
  if (fSDManager) delete fSDManager;
  if (fSubGroup)  delete fSubGroup;
  if (fName)  	  delete fName;
  if (fMother)	  delete fMother;
  if (fSolid)     delete fSolid;
  if (fLV)        delete fLV;
  if (fPV)        delete fPV;
  if (fSD)        delete fSD;
  
  if (!fDaughters.isEmpty()) fDaughters.clearAndDestroy();
  if (fCanDeleteUserLimits) delete fLV->GetUserLimits();
  
}

//=====================================================================
//* OpenMaterialStore -------------------------------------------------

J4VMaterialStore* J4VDetectorComponent::OpenMaterialStore()
{
	return 0;
}

//=====================================================================
//* GetNewName --------------------------------------------------------

G4String J4VDetectorComponent::GetNewName(G4String& middlename, // component class name
				 	   G4int firstname,      // I'm (firsthame)-th child
				 	   G4int width, 	 // field width for firstname 
				  	   G4String* parentname) // parent component name
{
  G4String* name;
  if (parentname) {
     name = new G4String(*parentname);
     *name += ":";
  } else {
     name = new G4String("");
  }
  *name += middlename;
  if (width){
     char tempstring[1024];
     char format[1024];
     sprintf(format,"%%%d.%dd",width,width);
     sprintf(tempstring,format,firstname);
     *name += tempstring;
  }
  G4String tempname(*name);
  delete name;
  return tempname;
}


//=====================================================================
//* OutputAll ---------------------------------------------------------

void J4VDetectorComponent::OutputAll(G4HCofThisEvent* HCTE)
{
   G4int ndaughters = fDaughters.entries();

#if 0
   G4cerr << "J4VDetectorComponent:OutputAll" << G4endl;
   G4cerr << "I am " << fName <<" . I have " << ndaughters 
          << " children. " << G4endl;
#endif
   
   G4int i;
   
   for( i=0; i<ndaughters; i++) 
   {      	
     if(fDaughters[i]->GetSD()) 
     {
        fDaughters[i]->GetSD()->OutputAll(HCTE);
     }
     
     fDaughters[i]->OutputAll(HCTE);
     
   }

}

//=====================================================================
//* SetDaughter ---------------------------------------------------------
void J4VDetectorComponent::SetDaughter(J4VDetectorComponent* daughter)
{
   fDaughters.append(daughter);
}

//=====================================================================
//* SetSD -------------------------------------------------------------

void J4VDetectorComponent::SetSD(J4VSensitiveDetector* sd)
{
  fSD = sd;
  if(fIsOn){
    if(!fLV) Assemble(); 
    fSDManager-> AddNewDetector(fSD);
    fLV-> SetSensitiveDetector(fSD);
  }
}

//=====================================================================
//* SetName --------------------------------------------------------

void J4VDetectorComponent::SetName(G4String& name,
				   G4int     nbrothers,
				   G4int     me,
				   J4VDetectorComponent *parent)
{
  if (nbrothers > 1) {
    G4int width = (G4int)log10(nbrothers + 0.1) + 1;   
    if (parent) {
      fName = GetNewName(name, me, width, &(parent->GetName()));
    } else {
      fName = GetNewName(name, me, width);
    }
  } else {
    if(parent){
      fName = GetNewName(name, 0, 0, &(parent->GetName()));
    }else  {
      fName = GetNewName(name);
    }
  } 
  
  G4cerr << " My name is " << fName << G4endl;
}

//=====================================================================
//* MakeLVWith --------------------------------------------------------

void J4VDetectorComponent::MakeLVWith( G4Material* material, G4UserLimits* ulp)
{
  fLV = new G4LogicalVolume(fSolid, material, fName); 
   
  if (!ulp) {
  	ulp = new G4UserLimits(_USTEPMAX_ , _UTRAKMAX_ , _UTIMEMAX_ ,
  			       _UEKINMIN_ , _URABGMIN_ );
  	fCanDeleteUserLimits = TRUE;
  }
  
  fLV->SetUserLimits(ulp);
  	
}

//=====================================================================
//* PaintLV   --------------------------------------------------------

void J4VDetectorComponent::PaintLV(G4bool visivility, const G4Color& color)
{
    G4VisAttributes* visatt
             = new G4VisAttributes(visivility, color);
    fLV-> SetVisAttributes(visatt);   
}

//=====================================================================
//* OrderNewTubs ------------------------------------------------------

void J4VDetectorComponent::OrderNewTubs( G4double rmin,
  					 G4double rmax,
  					 G4double len,
  					 G4double totalphi,
  					 G4double endcaphalfthickness,
  					 G4double endcaprmin,
  					 G4double sphi )
{
  G4double dphi;
  
#if 0
  if (fNClones == 1) {
    dphi = totalphi/fNClones;
  } else {	
    dphi = (totalphi/fNClones) - kRadTolerance;
  }
  
  rmin += kCarTolerance;
  rmax -= kCarTolerance;
#else  

  if (fNClones == 0) {
  	G4cerr << " J4VDetectorComponent::OrderNewTubs:"
  	       << "fNClones must be greater than 0 !! " << G4endl;
  }
  dphi = totalphi/fNClones;
#endif  
  
  if (sphi == DBL_MAX) {   
    sphi = (-dphi/2.);
  }
  
    
  if (!endcaphalfthickness) { 
  
    fSolid = new G4Tubs( fName, rmin, rmax, len, sphi, dphi );	
      
  } else { 
    
    if (endcaprmin == DBL_MAX) endcaprmin = rmin;
    
    G4cerr << "Endcap rmin = " << endcaprmin << G4endl;
    
    G4ThreeVector lefttranslation ( 0., 0., len - endcaphalfthickness);
    G4RotationMatrix rotation;
    G4Transform3D lefttransform   (rotation, lefttranslation);
    
    G4String cylindername(fName);
    cylindername += ".Cylinder";
    G4cerr << "Make " << cylindername << G4endl;
    G4double cylinderlen = len - endcaphalfthickness * 2;
    G4Tubs*  cylinder    
    	= new G4Tubs( cylindername, rmin, rmax, cylinderlen, sphi, dphi );

    G4String leftendcapname(fName);
    leftendcapname += ".LeftEndcap";
    G4cerr << "Make " << leftendcapname << G4endl;
    G4Tubs* leftendcap 
    	= new G4Tubs( leftendcapname, endcaprmin, rmax, endcaphalfthickness, sphi, dphi );
    
    G4String cupname(fName);
    cupname += ".Cup";
    G4cerr << "Make " << cupname << G4endl;
    J4UnionSolid*  cup = new J4UnionSolid ( cupname, cylinder, leftendcap, lefttransform );
        
    G4String rightendcapname(fName);
    rightendcapname += ".RightEndcap";
    G4cerr << "Make " << rightendcapname << G4endl;
    G4Tubs* rightendcap 
    	= new G4Tubs( rightendcapname, endcaprmin, rmax, endcaphalfthickness, sphi, dphi );
    
    G4ThreeVector    righttranslation ( 0., 0., -(len - endcaphalfthickness));
    G4Transform3D    righttransform   (rotation, righttranslation);
    G4cerr << "Make " << fName << G4endl;
    fSolid = new J4UnionSolid ( fName, cup, rightendcap, righttransform );    
    
  }
  
}

//=====================================================================
//* SetPVPlacement ----------------------------------------------------

void J4VDetectorComponent::SetPVPlacement(G4RotationMatrix *pRot,
                                          const G4ThreeVector &tlate)
{
  G4LogicalVolume*  motherLV = (fMother ? fMother->GetLV() : 0);
  
  fPV =  new G4PVPlacement(pRot,		// Rotation matrix
    		     	   tlate, 		// Three vector of center position
    			   fLV, 		// Logical volume
    			   fName,	 	// Name of Physical volume
    		           motherLV,  		// Mother Logical
    			   FALSE,		// pMany
    			   0      ); 		// pCopyNo	
}  

//=====================================================================
//* SetPVReplica   ----------------------------------------------------

void J4VDetectorComponent::SetPVReplica(const EAxis pAxis, G4double step,
  			                const G4double offset)
{ 
  G4LogicalVolume*  motherLV = (fMother ? fMother->GetLV() : 0);
  
  G4double doffset = offset;
  if ( pAxis == kPhi )  doffset -= 0.5*step;
  
  fPV = new G4PVReplica( fName,	        	// Name of Physical Volume
  			 fLV,			// Logical Volume
  		         motherLV,		// Mother Logical 
  			 pAxis,			// Direction of replication
  			 GetNClones(),		// Number of replicas
  			 step,			// Distance between adjacent clones
  			 doffset );		// Position of the first
}

   
   






